import jsgf
import json
import random
from collections import defaultdict
import jmespath
import re


class Rules_DST():

    def __init__(self):
        self.state = json.load(open('data.json'))

    def update_user(self, user_acts=None):
        try:
            for intent, domain, slot, value in user_acts:
                domain = domain.lower()
                intent = intent.lower()
                slot = slot.lower()
                if intent == 'start_conversation':
                    continue

                elif domain not in self.state['belief_state']:
                    continue


                elif 'inform' in intent:
                    if slot == 'name':
                        x, price = if_in_database(value)
                        self.state['belief_state']['order']['price'] += int(price)
                    if (slot == 'inform'):
                        continue

                    if (domain in slot):
                        slot.replace(domain + "/", '')

                    domain_dic = self.state['belief_state'][domain]
                    if slot in domain_dic:
                        self.state['belief_state'][domain][slot] = value


                elif intent == 'request':
                    if domain not in self.state['request_state']:
                        self.state['request_state'][domain] = {}
                    if slot not in self.state['request_state'][domain]:
                        self.state['request_state'][domain][slot] = 0
                    else:
                        self.state['request_state'][domain][slot] = value

                elif intent == 'start_conversation':
                    self.state["user_action"].append([intent, domain, slot, value])
                    continue

                elif intent == 'end_conversation':
                    self.state = json.load(open('data.json'))

                self.state["user_action"].append([intent, domain, slot, value])
        except:
            if user_acts[0] == 'end_conversation':
                self.state = json.load(open('data.json'))
        return self.state

class DP():
    def __init__(self):
        with open('database.json', encoding='utf-8-sig') as json_file:
            self.db = json.load(json_file)

    def predict(self, state):
        self.results = []
        system_action = defaultdict(list)
        user_action = defaultdict(list)
        system_acts = []
        for idx in range(len(state['user_action'])):
            intent, domain, slot, value = state['user_action'][idx]
            user_action[(domain, intent)].append((slot, value))

        for user_act in user_action:
            system_acts.append(self.update_system_action(user_act, user_action, state, system_action))
        state['system_action'] = system_acts
        return system_acts[-1]

    def update_system_action(self, user_act, user_action, state, system_action):

        domain, intent = user_act

        # Reguła 3
        if intent == 'end_conversation':
            return None

        constraints = [(slot, value) for slot, value in state['belief_state'][domain].items() if value != '']

        # Reguła 1
        if intent == 'request':
            if len(self.results) == 0:
                system_action[(domain, 'NoOffer')] = []
            else:
                for slot in user_action[user_act]:
                    if slot[0] in self.results[0]:
                        system_action[(domain, 'Inform')].append([slot[0], self.results[0].get(slot[0], 'unknown')])

        # Reguła 2
        elif intent == 'inform':
            if len(constraints) > 1:
                arg = f"{constraints[0]}".replace(f"\'{constraints[0][0]}\'", f"{constraints[0][0]}")
                arg = arg.replace("[", "").replace("]", "")
                for cons in constraints[1:]:
                    arg += f" && contains{cons}".replace(f"\'{cons[0]}\'", f"{cons[0]}").replace("[", "").replace("]",
                                                                                                                  "")
            else:
                arg = f"{constraints}".replace(f"\'{constraints[0]}\'", f"{constraints[0]}").replace("[", "").replace(
                    "]", "").replace("(\'", "(").replace("\',", ",")
            print(arg)
            print(jmespath.search(f"database.{domain}[?contains{arg} == `true` ]", self.db))
            self.results = jmespath.search(f"database.{domain}[?contains{arg} == `true` ]", self.db)
            if len(self.results) == 0:
                system_action[(domain, 'NoOffer')] = []
            else:
                system_action[(domain, 'Inform')].append(['Choice', str(len(self.results))])
                choice = self.results[0]

                if domain in ["food", "drink", "sauce"]:
                    system_action[(domain, 'Recommend')].append(['Name', choice['name']])
                elif domain in ["order", "booking", "payment"]:
                    system_action[(domain, 'Recommend')].append(['Type', choice['type']])
        return system_action

def nlu(utterance):
    matched = book_grammar.find_matching_rules(utterance)
    if not matched:
        matched = order_grammar.find_matching_rules(utterance)
    if not matched:
        matched = recommend_grammar.find_matching_rules(utterance)
    if not matched:
        matched = payment_grammar.find_matching_rules(utterance)
    if not matched:
        matched = bye_grammar.find_matching_rules(utterance)
    if not matched:
        matched = price_grammar.find_matching_rules(utterance)
    if matched:
        return get_dialog_act(matched[0], message)
    else:
        return {'act': 'null', 'slots': []}

def get_dialog_act(rule, message):
    slots = []
    get_slots(rule.expansion, slots)
    if rule.grammar.name == 'order':
        act = 'inform'
    elif rule.grammar.name == 'recommend' or rule.grammar.name == 'book':
        act = 'request'
    elif rule.grammar.name == 'payment':
        if slots == []:
            act = 'request'
        else:
            act = 'inform'
    elif rule.grammar.name == 'bye':
        act = 'end_conversation'
    elif rule.grammar.name == 'price':
        act = 'request'
        slots = [('price', 'amount')]
    if act == 'inform' and slots[0] == 'drink':
        slots[1] = message.split()[-1]
    return {'act': act, 'slots': slots}

def get_slots(expansion, slots):
    if expansion.tag != '':
        slots.append((expansion.tag, expansion.current_match))
        return

    for child in expansion.children:
        get_slots(child, slots)

    if not expansion.children and isinstance(expansion, jsgf.NamedRuleRef):
        get_slots(expansion.referenced_rule.expansion, slots)

def nlg(system_act):
    try:
        intent, domain, slot, value = system_act

        if intent == 'cant_help':
            return 'Nie rozumiem, możesz inaczej sformulowac swoja mysl'
        if domain == 'payment' and intent == 'inform':
            if slot == 'type':
                if value == 'blik' or value == 'gotówka' or value == 'karta' or value == 'karta restauracyjna' or value == 'bon upominkowy':
                    return f"Informuje, że {value} jest możliwa forma płatności w naszej restauracji."
                else:
                    return f"Informuje, że {value} nie jest możliwa forma płatności w naszej restauracji."
            if slot == 'vat':
                return f"Twój vat jest {value} w rachunek."
            if slot == 'amount':
                return f"Kwota Twojej płatności to {value}."
            else:
                return "Nie moge wykonać takiej operacji."
        elif domain == 'booking' and intent == 'inform':
            return f"Nasza restauracja nie umozliwia rezerwacji stolikow."
        elif domain == 'dish' and intent == 'inform':
            if slot == 'name':
                in_db, price = if_in_database(value)
                if in_db:
                    return f"Informuje, że dodałem {value} do zamowienia, koszt zwiekszyl sie o {price} zl."
                else:
                    return f"Informuje, że nie posiadamy {value} w naszej ofercie."

        elif domain == 'drink' and intent == 'inform':
            if slot == 'name':
                in_db, price = if_in_database(value)
                if in_db:
                    return f"Informuje, że dodałem {value} do zamowienia, koszt zwiekszyl sie o {price} zl."
                else:
                    return f"Informuje, że nie posiadamy {value} w naszej ofercie."

        elif domain == 'sauce' and intent == 'inform':
            if slot == 'size':
                if value == 'mała':
                    return f"Inforumuje, że sos o rozmiarze {value} jest serwowany w pojemniku plastikowym."
                elif value == 'duża':
                    return f"Inforumuje, że sos o rozmiarze {value} jest serwowany w szklanej misce."

            if slot == 'price_range':
                if value == 'niska':
                    return f"Inforumuje, {value} cena dotyczy sosów podawanych w plastikowym pojemniku."
                elif value == 'duża':
                    return f"Inforumuje, {value} cena dotyczy sosów podawanych w szklanej miseczce."

            if slot == 'name':
                if value in ["sos pomidorowy", "sos czosnkowy", "sos bbq", "sos tatarski", "sos Meksykanski", "sos aioli",
                             "sos śmietanowo-koperkowy", "sos pikantny", "sos curry", "sos hollandaise"]:
                    return f"Informuje, że posiadamy {value} w naszej ofercie."
                else:
                    return f"Informuje, że nie posiadamy {value} w naszej ofercie."

        elif intent == 'start_conversation':
            return random.choice(['Witaj, jestem asystentem dialogowym, dzieki któremu złożysz zamówienie w restuaracji.',
                                  'Cześć, w czym moge pomóc?', "Witam w usłudze, w czym moge pomóc?"])
        elif intent == 'affirm':
            return random.choice(['Tak.', 'Potwierdzam.', "Mozliwe."])
        elif intent == 'deny':
            return random.choice(['Nie.', 'Nie potwierdzam.', "Niemożliwe."])
        elif intent == 'request' and slot == "type":
            if domain == "dish":
                if value is not None:
                    if value in ["pizza", "pizze"]:
                        dish_type = "pizza"
                    if value in ["makaron", "makarony"]:
                        dish_type = "makaron"
                    if value in ["salatka", "salatke", "salatki"]:
                        dish_type = "salatka"
                    if value in ["deser", "desery"]:
                        dish_type = "deser"
                    if value in ["sniadania", "sniadanie"]:
                        dish_type = "sniadanie"
                    if value in ["danie", "dania", "potrawa", "potrawy", "potrawe"]:
                        food_all = ["pizza", "makaron", "salatka", "deser", "sniadanie"]
                        food_type = random.choice(food_all)
                    recommended = get_by_type(food_type)
                    recommended = random.choice(recommended)
                    return f"Rekomendowany posiłek to: {recommended}."
                else:
                    food_types = ["deser", "salatka", "makaron", "zupa", "danie", "pizza", "sniadanie"]
                    recommendations = get_by_type(random.choice(food_types))
                    return "Polecam " + str(random.choice(recommendations))
            if domain == "drink":
                if value is not None:
                    if value in ["sok", "soki"]:
                        drink_type = "sok"
                    if value in ["herbata", "herbaty", "herbate"]:
                        drink_type = "herbata"
                    if value in ["piwo", "piwa"]:
                        drink_type = "piwo"
                    if value in ["kawa", "kawe"]:
                        drink_type = "kawa"
                    if value in ["wino", "wina"]:
                        drink_type = "wino"
                    if value in ["napoj", "napoje"]:
                        drinks_all = ["sok", "herbata", "piwo", "kawa", "wino"]
                        drink_type = random.choice(drinks_all)
                    recommended = get_by_type(drink_type)
                    recommended = random.choice(recommended)
                    return f"Rekomendowany napój to: {recommended}."
                else:
                    drink_types = ["kawa", "herbata", "cola", "sok", "woda", "piwo", "wino"]
                    recommendations = get_by_type(random.choice(drink_types))
                    return "Polecam " + str(random.choice(recommendations))
            if domain == "sauce":
                return f"Rekomendowany sos to: {value}."
        elif intent == 'recommend' and slot == "type":
            if domain == "booking":
                return f"Rekomenduje nastepujaca rezerwacje stolika: {value}."
            if domain == "order":
                return f"Rekomenduje takie zamówienie: {value}."
            if domain == "payment":
                return f"Rekomendowana metoda płatności to: {value}."
    except:
        if system_act[0] == 'end_conversation':
            return random.choice(['Żegnam i dziekuje za skorzystanie z usługi.', 'Do widzenia.',
                              "Dziekuje za skorzystanie z usługi, do kolejnego razu."])
        if system_act[0] == 'request' and system_act[1] == 'price':
            curr_price = dst.state['belief_state']['order']['price']
            return f"Twoje zamówienie akutalnie kosztuje {curr_price} zl."
        if system_act[0] == 'request' and system_act[1] == 'day':
            return f"Przepraszamy, niestety nasza restauracja nie umozliwia rezerwacji stolikow."

def generate_dst_in(nlu_out):
    dst_in = []
    for key, val in nlu_out['slots']:
        temp = []
        temp.append(nlu_out['act'])
        if (key == 'type') and nlu_out['act'] == 'inform':
            temp.append('payment')
        temp.append(key)
        if (key == 'dish' or key == 'drink') and nlu_out['act'] == 'inform':
            temp.append('name')
        if (key == 'dish' or key == 'drink') and nlu_out['act'] == 'request':
            temp.append('type')
        temp.append(val)
        dst_in.append(temp)
    return dst_in

def get_by_type(type):
    with open('database.json') as file:
        data = json.load(file)

    list_of_type = []
    # get all values of a type
    for key in data["database"]:
        value = data["database"][key]
        for val in value:
            if val.get("type") == type:
                list_of_type.append(val.get("name"))
    return list_of_type

def if_in_database(name):
    with open('database.json') as file:
        data = json.load(file)

    for key in data["database"]:
        value = data["database"][key]
        for val in value:
            if val.get("name") == name:
                price = val.get("price")
                return True, price
    return False, 0

def get_by_name(name):
    with open('database.json') as file:
        data = json.load(file)

    for key in data["database"]:
        value = data["database"][key]
        for val in value:
            if str(val.get("name")) == name:
                return val
    return None

import HELP

#inicjalizacja
book_grammar = jsgf.parse_grammar_file('book.jsgf')
order_grammar = jsgf.parse_grammar_file('order.jsgf')
recommend_grammar = jsgf.parse_grammar_file('recommend.jsgf')
payment_grammar = jsgf.parse_grammar_file('payment.jsgf')
bye_grammar = jsgf.parse_grammar_file('bye.jsgf')
price_grammar = jsgf.parse_grammar_file('price.jsgf')

dishes = ["pizze peperoni", "spaghetti carbonara", "burger szefa", "burgera szefa",
            "pizze margherita", "spaghetti bolognese", "burger wegetariański", "burgera wegetariańskiego",
            "pizze capricciosa", "spaghetti aglio e olio", "burger z karmelizowaną cebulą", "burgera z karmelizowaną cebulą",
            "pizze quattro formaggi", "spaghetti alle vongole", "burger z grzybami i serem",
            "pizze hawajska", "spaghetti alla puttanesca", "burger z klasycznymi dodatkami", "burgera z klasycznymi dodatkami",
            "pizze funghi", "spaghetti al pesto", "burger z sosem BBQ", "burgera z sosem BBQ",
            "pizze diavola", "spaghetti alla marinara", "burger z krewetkami i awokado", "burgera z krewetkami i awokado",
            "pizze prosciutto e funghi", "spaghetti alla carbonara", "burger z jalapenos i serem", "burger z jalapeno i serem"]

drinks = ["cola", "cole", "lemoniada", "lemoniade"]

dst = Rules_DST()
dp = DP()
print("Witaj w restauracji pizzowo. Jestem asystentem, ktory ma ci pomoc zlozyc zamowienie. \nJesli nie jestes pewien jak ze mna rozmawiac wpisz help. \nW czym moge sluzyc?")

#input
while True:
    message = input("User: ")
    message = re.sub(r'\W+', ' ', message)
    message = message.lower()
    if message == 'help':
        HELP.get_help()
    else:
        nlu_out = nlu(message)
        dst_in = generate_dst_in(nlu_out)
        # dst_in = [['inform', 'drink', 'name', 'cola']]
        dst.update_user(dst_in)
        for dst_el in dst_in:
            print(nlg(dst_el))
        if dst_in == []:
            print(nlg(['cant_help', '', '', '']))
        elif nlu_out['act'] == 'end_conversation':
            break
        else:
            print('Czy moge jeszcze w czyms pomoc?')
# print(nlg(dst_in[0]))

